#line 79 "./lpsrc/flx_gc.pak"
#ifndef __FLX_GC_H__
#define __FLX_GC_H__

#include <cstdlib>
#include "flx_gc_config.hpp"

// we use an STL set to hold the collection of roots
#include <set>

namespace flx {
namespace gc {
namespace generic {
// Here are the types we refer to:

struct GC_EXTERN frame_t;      // the type of all collectable objects
struct GC_EXTERN gc_shape_t;   // the shape of collectable objects
struct GC_EXTERN collector_t;  // the collector itself
struct GC_EXTERN allocator_t;  // the collector itself

enum gc_shape_flags_t {
  gc_flags_default    = 0,            //< collectable and mobile
  gc_flags_immobile   = 1,            //< cannot be moved
  gc_flags_persistent = 2             //< cannot be deallocated
};

/// Describes runtime object shape.
struct GC_EXTERN gc_shape_t
{
  gc_shape_t *next_shape;         ///< pointer to next shape in list or NULL
  char const *cname;              ///< C++ typename
  std::size_t count;              ///< array element count
  std::size_t amt;                ///< bytes allocated
  void (*finaliser)(collector_t*, void*);  ///< finalisation function
  std::size_t n_offsets;          ///< number of offsets
  std::size_t *offsets;           ///< actual offsets
  gc_shape_flags_t flags;         ///< flags
  // convenience constructor
  gc_shape_t(
    gc_shape_t *ns,
    char const *cn,
    std::size_t count_a,
    std::size_t amt_a,
    void (*finaliser_a)(collector_t*, void*),
    std::size_t n_offsets_a,
    std::size_t *offsets_a
  );
  gc_shape_t(
    gc_shape_t *ns,
    char const *cn,
    std::size_t count_a,
    std::size_t amt_a,
    void (*finaliser_a)(collector_t*, void*),
    std::size_t n_offsets_a,
    std::size_t *offsets_a,
    gc_shape_flags_t flags_a
  );
};
#line 160 "./lpsrc/flx_gc.pak"
template<class T>
void std_finaliser(collector_t*, void *t)
{
  static_cast<T*>(t) -> ~T();
}

#line 169 "./lpsrc/flx_gc.pak"

/// Allocator abstraction.

struct allocator_t {
  bool debug;
  allocator_t():debug(false){}
  virtual void *allocate(std::size_t)=0;
  virtual void deallocate(void *)=0;
  virtual void *reallocate(void *, std::size_t)=0;
  virtual ~allocator_t(){};
  void set_debug(bool d){debug=d;}
};

#line 185 "./lpsrc/flx_gc.pak"

/// Collector abstraction.
struct GC_EXTERN collector_t
{
  bool debug;
  void set_debug(bool d){debug=d;}
  collector_t();
  virtual ~collector_t(){}

#line 197 "./lpsrc/flx_gc.pak"
  unsigned long get_allocation_count()const {
    return v_get_allocation_count();
  }

  unsigned long get_root_count()const {
    return v_get_root_count();
  }

  unsigned long get_allocation_amt()const {
    return v_get_allocation_amt();
  }

#line 213 "./lpsrc/flx_gc.pak"
  void *allocate(gc_shape_t *shape, unsigned long x) {
    return v_allocate(shape,x);
  }

  void deallocate(frame_t *fp) {
    v_deallocate(fp);
  }

#line 224 "./lpsrc/flx_gc.pak"
  unsigned long collect() {
    return v_collect();
  }

#line 231 "./lpsrc/flx_gc.pak"
  void add_root(void *memory) {
    v_add_root(memory);
  }

  void remove_root(void *memory) {
    v_remove_root(memory);
  }

#line 253 "./lpsrc/flx_gc.pak"
  void compact(bool closed) {
    v_compact(closed);
  }

#line 260 "./lpsrc/flx_gc.pak"
  void check() {
    v_check();
  }

private:
  virtual unsigned long v_get_allocation_count()const=0;
  virtual unsigned long v_get_root_count()const=0;
  virtual unsigned long v_get_allocation_amt()const=0;
  virtual void *v_allocate(gc_shape_t *shape, unsigned long)=0;
  virtual void v_deallocate(frame_t *fp)=0;
  virtual unsigned long v_collect()=0;
  virtual void v_add_root(void *memory)=0;
  virtual void v_remove_root(void *memory)=0;
  virtual void v_compact(bool closed)=0;
  virtual void v_check()=0;

#line 280 "./lpsrc/flx_gc.pak"
  void operator=(collector_t const&);
  collector_t(collector_t const&);
};


#line 292 "./lpsrc/flx_gc.pak"
void GC_EXTERN destroy(void *b);

#line 306 "./lpsrc/flx_gc.pak"
void GC_EXTERN _init_ptr(void **a, void *b);
void GC_EXTERN _set_ptr(void **a, void *b);
void GC_EXTERN _release_ptr(void **a);
void GC_EXTERN _destroy_ptr(void **a);

template<class T>
void init_ptr(T **a, T *b)
{
  _init_ptr
  (
    reinterpret_cast<void**>(a),
    reinterpret_cast<void*>(b)
  );
}

template<class T>
void set_ptr(T **a, T *b)
{
  _set_ptr
  (
    reinterpret_cast<void**>(a),
    reinterpret_cast<void*>(b)
  );
}

template<class T>
void release_ptr(T **a)
{
  _release_ptr
  (
    reinterpret_cast<void**>(a)
  );
}

template<class T>
void destroy_ptr(T **a)
{
  _destroy_ptr
  (
    reinterpret_cast<void**>(a)
  );
}

#line 355 "./lpsrc/flx_gc.pak"
GC_EXTERN void set_used(void *memory, unsigned long);
GC_EXTERN void incr_used(void *memory, unsigned long);
GC_EXTERN unsigned long get_used(void *memory);
GC_EXTERN unsigned long get_count(void *memory);
GC_EXTERN void *create_empty_array(
  collector_t &collector,
  gc_shape_t &shape,
  unsigned long count
);

}}} // end namespaces

#line 377 "./lpsrc/flx_gc.pak"
/// Allocate collectable object
GC_EXTERN void *operator new
(
  std::size_t,
  flx::gc::generic::collector_t &,
  flx::gc::generic::gc_shape_t &
);
#endif

