#line 1230 "./lpsrc/flx_rtl.pak"
#include "flx_rtl.hpp"
#include <cstdio>
#include <cassert>

namespace flx { namespace rtl {

// ********************************************************
// con_t implementation
// ********************************************************

con_t::con_t() : pc(0), p_svc(0), _caller(0) {
#if FLX_DEBUG_CONT
 fprintf(stderr,"Constructing %p\n",this);
#endif
}
con_t::~con_t(){
#if FLX_DEBUG_CONT
  fprintf(stderr,"Destroying %p\n",this);
#endif
}

// ********************************************************
// slist implementation
// ********************************************************

// PRIVATE to the slist_t class!
struct slist_node_t {
  slist_node_t *next;
  void *data;
  slist_node_t(slist_node_t *n, void *d) : next(n), data(d) {}
};

//OFFSETS for slist_node_t
static std::size_t slist_node_offsets[2]={
    offsetof(slist_node_t,next),
    offsetof(slist_node_t,data)
};

flx::gc::generic::gc_shape_t slist_node_ptr_map(
  NULL,
  "slist_node_t",
  1,sizeof(slist_node_t),
  0, // no finaliser,
  2,
  slist_node_offsets
);


slist_t::slist_t(gc::generic::collector_t *_gc) : gc (_gc), head(0) {}
slist_t::slist_t(slist_t const &r) : gc (r.gc), head(r.head) {}

bool slist_t::isempty()const { return head == 0; }

void slist_t::push(void *data)
{
  head = new(*gc,slist_node_ptr_map) slist_node_t(head,data);
}

// note: never fails, return NULL pointer if the list is empty
void *slist_t::pop()
{
  if(head) {
    void *data = head->data;
    head=head->next;
    return data;
  }
  else return 0;
}

//OFFSETS for slist_t
static std::size_t slist_offsets[1]={
    offsetof(slist_t,head)
};

flx::gc::generic::gc_shape_t slist_ptr_map(
  &slist_node_ptr_map,
  "slist_t",
  1,sizeof(slist_t),
  0, // no finaliser
  1,
  slist_offsets
);

// ********************************************************
// _root_ptr_t implementation
// ********************************************************

_root_ptr_t::_root_ptr_t() : next(0), prev(0), data(0) {}
_root_ptr_t::_root_ptr_t(void *x) : next(0), prev(0), data(x) {}
_root_ptr_t::~_root_ptr_t() { erase(); }

void _root_ptr_t::insert_after(_root_ptr_t *a) {
  prev = a;
  next = prev->next;
  prev->next = this;
  if(next)next->prev = this;
  data = next->data;
}

_root_ptr_t::_root_ptr_t(_root_ptr_t const &a) {
  insert_after(const_cast<_root_ptr_t*>(&a));
}

void _root_ptr_t::operator=(_root_ptr_t const &a) {
  if (&a != this)
  {
    erase();
    insert_after(const_cast<_root_ptr_t*>(&a));
  }
}

//OFFSETS for root_ptr_t
static std::size_t _root_ptr_offsets[3]={
    offsetof(_root_ptr_t,next),
    offsetof(_root_ptr_t,prev),
    offsetof(_root_ptr_t,data)
};

flx::gc::generic::gc_shape_t _root_ptr_ptr_map(
  &slist_ptr_map,
  "_root_ptr_t",
  1,sizeof(_root_ptr_t),
  0, // no finaliser
  1,
  _root_ptr_offsets
);

void _root_ptr_t::erase() {
  next->prev = prev;
  prev->next = next;
  next = 0; prev = 0;
  data = 0;
}

// ********************************************************
// fthread_t implementation
// ********************************************************

fthread_t::fthread_t() : cc(0) {}
fthread_t::fthread_t(con_t *a) : cc(a) {}

// uncopyable object but implementation needed for linker
fthread_t::fthread_t(fthread_t const&){ assert(false); }
void fthread_t::operator=(fthread_t const&){ assert(false); }

void fthread_t::kill() { cc = 0; }

_uctor_ *fthread_t::get_svc()const { return cc?cc->p_svc:0; }

_uctor_ *fthread_t::run() {
  if(!cc) return 0; // dead
restep:
  cc->p_svc = 0;
step:
  //fprintf(stderr,"cc=%p->",cc);
  try { cc = cc->resume(); }
  catch (con_t *x) { cc = x; }

  //fprintf(stderr,"->%p\n",cc);
  if(!cc) return 0; // died

  if(cc->p_svc)
  {
    switch(cc->p_svc->variant)
    {
      case svc_get_fthread:
        **(fthread_t***)(cc->p_svc->data) = this;
        goto restep;      // handled

      //case svc_yield:
      //  goto restep;

      // we don't know what to do with the request,
      // so pass the buck to the driver
      default:
        return cc->p_svc;
    }
  }
  goto step;
}


//OFFSETS for fthread_t
static std::size_t _fthread_offsets[1]={
    offsetof(fthread_t,cc)
};

flx::gc::generic::gc_shape_t _fthread_ptr_map(
  &_root_ptr_ptr_map,
  "fthread_t",
  1,sizeof(fthread_t),
  0,
  1,
  _fthread_offsets,
  gc::generic::gc_flags_immobile
);

// ********************************************************
// schannel_t implementation
// ********************************************************

schannel_t::schannel_t (gc::generic::collector_t *gc) :
  waiting_to_read(0), waiting_to_write(0)
{
  waiting_to_read = new (*gc, slist_ptr_map) slist_t(gc);
  waiting_to_write = new (*gc, slist_ptr_map) slist_t(gc);
}

// uncopyable object but implementation needed for linker
schannel_t::schannel_t(schannel_t const&) { assert(false); }
void schannel_t::operator=(schannel_t const&) { assert(false); }

void schannel_t::push_reader(fthread_t *r)
{
  waiting_to_read->push(r);
}

void schannel_t::push_writer(fthread_t *w)
{
  waiting_to_write->push(w);
}

fthread_t *schannel_t::pop_reader()
{
  return (fthread_t*)waiting_to_read->pop();
}

fthread_t *schannel_t::pop_writer()
{
  return (fthread_t*)waiting_to_write->pop();
}

//OFFSETS for schannel_t
static std::size_t schannel_offsets[2]={
    offsetof(schannel_t,waiting_to_read),
    offsetof(schannel_t,waiting_to_write)
};

flx::gc::generic::gc_shape_t schannel_ptr_map(
  &_fthread_ptr_map,
  "schannel_t",
  1,sizeof(schannel_t),
  0, // no finaliser
  2,
  schannel_offsets
);

// ********************************************************
// _ref_ implementation
// ********************************************************

void _ref_::operator = (_ref_ const& r)
{
  if (frame != r.frame || offset != r.offset)
  {
    frame = r.frame;
    offset = r.offset;
  }
}

//OFFSETS for _ref_
static std::size_t _ref_offsets[1]={
    offsetof(_ref_,frame)
};

flx::gc::generic::gc_shape_t _ref_ptr_map(
  &schannel_ptr_map,
  "_ref_",
  1,sizeof(_ref_),
  0, // no finaliser
  1,
  _ref_offsets
);

// ********************************************************
// _uctor_ implementation
// ********************************************************

//OFFSETS for _uctor_
static std::size_t _uctor_offsets[1]= {
  offsetof(_uctor_,data)
};

flx::gc::generic::gc_shape_t _uctor_ptr_map (
  &_ref_ptr_map,
  "_uctor_",
  1,
  sizeof(_uctor_),
  0,
  1,
  _uctor_offsets
);

// ********************************************************
// int implementation
// ********************************************************

flx::gc::generic::gc_shape_t _int_ptr_map (
  &_uctor_ptr_map,
  "int",
  1,
  sizeof(int),
  0,
  0,
  0
);

// ********************************************************
// unit implementation
// ********************************************************

flx::gc::generic::gc_shape_t unit_ptr_map (
  &_int_ptr_map,
  "unit",
  1,
  sizeof(unit),
  0,
  0,
  0
);

// ********************************************************
// standard exceptions -- implementation
// ********************************************************
flx_exception_t::~flx_exception_t(){}

flx_exec_failure_t::flx_exec_failure_t(std::string f, std::string o, std::string w) :
  filename(f),
  operation(o),
  what(w)
{}

flx_exec_failure_t::~flx_exec_failure_t(){}

flx_range_srcref_t::flx_range_srcref_t() :
    filename(""),startline(0),startcol(0),endline(0),endcol(0){}
flx_range_srcref_t::flx_range_srcref_t(char *f,int sl, int sc, int el, int ec) :
    filename(f),startline(sl),startcol(sc),endline(el),endcol(ec){}

flx_halt_t::flx_halt_t(flx_range_srcref_t ff, char *cf, int cl, std::string r) :
   reason(r), flx_loc(ff), cxx_srcfile(cf), cxx_srcline(cl) {}
flx_halt_t::~flx_halt_t(){}

flx_match_failure_t::flx_match_failure_t(flx_range_srcref_t ff, char *cf, int cl) :
   flx_loc(ff), cxx_srcfile(cf), cxx_srcline(cl) {}
flx_match_failure_t::~flx_match_failure_t(){}

flx_assert_failure_t::flx_assert_failure_t(flx_range_srcref_t ff, char *cf, int cl) :
   flx_loc(ff), cxx_srcfile(cf), cxx_srcline(cl) {}
flx_assert_failure_t::~flx_assert_failure_t(){}

flx_assert2_failure_t::flx_assert2_failure_t(flx_range_srcref_t ff, flx_range_srcref_t ff2, char *cf, int cl) :
   flx_loc(ff), flx_loc2(ff2), cxx_srcfile(cf), cxx_srcline(cl) {}
flx_assert2_failure_t::~flx_assert2_failure_t(){}

flx_range_failure_t::flx_range_failure_t(flx_range_srcref_t ff, char *cf, int cl) :
   flx_loc(ff), cxx_srcfile(cf), cxx_srcline(cl) {}
flx_range_failure_t::~flx_range_failure_t(){}

flx_switch_failure_t::~flx_switch_failure_t(){}

long range_check (long l, long x, long h, flx_range_srcref_t sref, char *cf, int cl)
{
  if (x>=l && x<h) return x;
  throw flx::rtl::flx_range_failure_t (sref,cf,cl);
}


}}
